/*
 * File-specific pager functionality.
 *
 * Copyright (C) 2021, 2022, 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <fsclient/file.h>  /* file_region_flags */
#include <systypes/fcntl.h>

#include "copied_page_mapper.h"
#include "file_pager.h"
#include "mapped_file_object_server.h"
#include "resource_server.h"



/* Initialise a pager for a file with a unique file identifier, file provider,
   opening flags and a file registry. The provider offers a shared page mapper
   for moderating access to loaded pages. */

FilePager::FilePager(fileid_t fileid, FileProvider *provider, flags_t flags)
: Pager(provider->mapper(), file_region_flags(flags)),
  _provider(provider), _flags(flags), fileid(fileid)
{
}

ipc_server_default_config_type FilePager::config()
{
    return config_MappedFileObject;
}



/* Close the pager, removing the provider from the registry if appropriate. */

void FilePager::close()
{
    /* Notify other users of the file and unsubscribe. */

    _provider->notify_others(_notifier, NOTIFY_PEER_CLOSED, NOTIFY_VALUES_NULL);
    unsubscribe(_notifier);

    /* Detach the pager, potentially removing the file provider. */

    _provider->registry()->detach(fileid, _provider);
}



/* File-specific operations. */

long FilePager::flush(offset_t populated_size, offset_t *size)
{
    long err = Pager::flush(populated_size, size);

    if (_resized)
    {
        _provider->notify_others(_notifier, NOTIFY_CONTENT_AVAILABLE, NOTIFY_VALUES_NULL);
        _resized = false;
    }

    return err;
}

long FilePager::reopen(flags_t flags, offset_t *size, l4_cap_idx_t *file,
                       object_flags_t *object_flags)
{
    Resource *resource;
    long err = _provider->make_resource(flags, size, object_flags, &resource);

    if (err)
        return err;

    _provider->notify_all(NOTIFY_FILE_OPENED, NOTIFY_VALUES_NULL);

    return ResourceServer(resource).start_thread(file);
}

long FilePager::resize(offset_t *size)
{
    offset_t old_size = get_data_size();
    long err = Pager::resize(size);

    /* If the size has changed, notify other users, but only after data has been
       flushed. */

    if (old_size < get_data_size())
        _resized = true;

    return err;
}

long FilePager::mmap(offset_t position, offset_t length,
                     offset_t start_visible, offset_t end_visible,
                     offset_t *start_pos, offset_t *end_pos, offset_t *size)
{
    /* Set the limits of the paged region. */

    long err = Pager::mmap(position, length, start_visible, end_visible, start_pos,
                           end_pos, size);

    if (err)
        return err;

    /* Impose copy-on-write semantics where appropriate. */

    if ((_mapper != _mapper_copied) && copy_on_write())
    {
        _mapper_copied = new CopiedPageMapper(_mapper);
        _mapper = _mapper_copied;
    }

    return L4_EOK;
}

/* Return whether the pager should employ copy-on-write semantics. */

bool FilePager::copy_on_write()
{
    return (_flags == O_RDONLY) && (_map_flags & L4RE_DS_F_W);
}



/* Generic pager operations. */

long FilePager::map(offset_t offset, map_address_t hot_spot, map_flags_t map_flags, l4_snd_fpage_t *region)
{
    return Pager::map(offset, hot_spot, map_flags, region);
}



/* Subscribe to notifications.
   Readers can subscribe to new data (flush) and file closed events.
   Writers can subscribe to file closed events. */

long FilePager::subscribe(l4_cap_idx_t notifier, notify_flags_t flags)
{
    /* A single notifier is recorded so that it may be unsubscribed when the
       file is closed. */

    if ((notifier != _notifier) && l4_is_valid_cap(_notifier))
        unsubscribe(_notifier);

    _notifier = notifier;
    return _provider->subscribe(notifier, flags);
}

/* Unsubscribe from notifications. */

long FilePager::unsubscribe(l4_cap_idx_t notifier)
{
    return _provider->unsubscribe(notifier);
}

// vim: tabstop=4 expandtab shiftwidth=4
